/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.shardingsphere.sharding.distsql.segment.table;

import lombok.Getter;
import lombok.RequiredArgsConstructor;
import org.apache.shardingsphere.distsql.segment.DistSQLSegment;

import java.util.Arrays;
import java.util.Collection;
import java.util.stream.Collectors;

/**
 * Table reference rule segment.
 */
@RequiredArgsConstructor
@Getter
public final class TableReferenceRuleSegment implements DistSQLSegment {
    
    private final String name;
    
    private final String reference;
    
    /**
     * Get table names.
     *
     * @return table names
     */
    public Collection<String> getTableNames() {
        return Arrays.stream(reference.split(",")).map(String::trim).collect(Collectors.toList());
    }
}
